package com.gihon.component.rbac.service.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.StringUtils;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.gihon.component.entity.BaseEntity;
import com.gihon.component.exception.BusinessException;
import com.gihon.component.rbac.entity.GihonButton;
import com.gihon.component.rbac.entity.GihonRole;
import com.gihon.component.rbac.entity.GihonUserRole;
import com.gihon.component.rbac.listener.CustomePushlisher;
import com.gihon.component.rbac.listener.CustomerEvent;
import com.gihon.component.rbac.mapper.GihonRoleMapper;
import com.gihon.component.rbac.service.GihonRoleService;
import com.gihon.component.rbac.vo.RoleButtonReq;
import com.gihon.component.rbac.vo.RoleListReq;
import com.gihon.component.response.SelectVal;

/**
 * 直接继承没有多实现处理
 * 
 * @author baihe
 *
 */
@Service
public class GihonRoleServiceImpl extends ServiceImpl<GihonRoleMapper, GihonRole> implements GihonRoleService {

    @Autowired
    private GihonUserRoleServiceImpl gihonUserRoleService;

    @Autowired
    private GihonRoleButtonServiceImpl gihonRoleButtonService;

    @Autowired
    private GihonButtonServiceImpl gihonButtonService;

    @Autowired
    private GihonMenuServiceImpl gihonMenuService;

    @Autowired
    protected CustomePushlisher customePushlisher;

    /**
     * 获取用户的角色列表
     */
    public List<GihonRole> getRoleList(Long userId) {
        List<Long> roleList = gihonUserRoleService.roleListByUserId(userId);
        return this.getRoleListByIds(roleList);
    }

    /**
     * 按照角色IDS获取角色列表
     */
    public List<GihonRole> getRoleListByIds(List<Long> ids) {
        if (ids.isEmpty()) {
            return new ArrayList<>(0);
        }
        return this.lambdaQuery().in(GihonRole::getId, ids).list();
    }

    /**
     * 前端分页获取角色列表
     */
    public IPage<GihonRole> getRolelistPage(RoleListReq companyListReq) {
        Page<GihonRole> page = new Page<>(companyListReq.getPageIndex(), companyListReq.getPageSize());
        return this.lambdaQuery()
            .like(StringUtils.hasText(companyListReq.getRoleName()), GihonRole::getRoleName,
                companyListReq.getRoleName())
            .like(StringUtils.hasText(companyListReq.getRoleCode()), GihonRole::getRoleCode,
                companyListReq.getRoleCode())
            .eq(companyListReq.getCompanyId() != null, GihonRole::getCompanyId, companyListReq.getCompanyId())
            .page(page);
    }

    /**
     * 新增角色
     */
    public boolean saveRole(GihonRole role) {
        int count = this.lambdaQuery().eq(GihonRole::getRoleName, role.getRoleName()).or()
            .eq(GihonRole::getRoleCode, role.getRoleCode()).count();
        if (count > 0) {
            throw new BusinessException("角色名称或编码重复");
        }
        role.setState(0);
        return this.save(role);
    }

    /**
     * 更新角色
     */
    public boolean updateRole(GihonRole role) {
        GihonRole roleOld = this.getById(role.getId());
        if (roleOld == null) {
            throw new BusinessException("角色不存在");
        }
        int count = this.lambdaQuery().ne(GihonRole::getId, role.getId()).and(p -> {
            p.eq(GihonRole::getRoleName, role.getRoleName()).or().eq(GihonRole::getRoleCode, role.getRoleCode());
        }).count();

        if (count > 0) {
            throw new BusinessException("角色名称或编码重复");
        }
        GihonRole roleUpdate = new GihonRole();
        roleUpdate.setId(role.getId());
        roleUpdate.setRoleName(role.getRoleName());
        roleUpdate.setRoleCode(role.getRoleCode());
        roleUpdate.setRoleDescription(role.getRoleDescription());
        return this.updateById(roleUpdate);

    }

    /**
     * 移除一个角色
     */
    @Transactional(rollbackFor = Exception.class)
    public boolean removeRole(Long roleId) {
        GihonRole roleOld = this.getById(roleId);
        if (roleOld == null) {
            throw new BusinessException("角色不存在");
        }
        // 异步通知
        gihonRoleButtonService.removeButtonByRoleId(roleId, null);
        
        LambdaQueryWrapper<GihonUserRole> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(GihonUserRole::getRoleId, roleId);
        gihonUserRoleService.remove(wrapper);

        CustomerEvent event = new CustomerEvent(roleId);
        event.setType(2);
        event.setId(roleId);
        customePushlisher.publish(event);

        return this.removeById(roleId);
    }

    /**
     * 获取在某个模块下的按钮权限 如果是缓存的时候需要异步通知
     */
    public List<GihonButton> getButtonByRoleModule(Long roleId, Long moduleId) {
        List<Long> menuIdList =
            gihonMenuService.getAllMenu(moduleId, true).stream().map(BaseEntity::getId).collect(Collectors.toList());
        List<GihonButton> btnList = gihonButtonService.getMenuButton(menuIdList, false);
        List<Long> btnCheckedList = gihonRoleButtonService.getButtonListByRole(roleId);
        btnList.forEach(b -> b.setChecked(btnCheckedList.contains(b.getId())));
        return btnList;
    }

    /**
     * 设置角色在某个模块下的按钮权限
     */
    @Transactional(rollbackFor = Exception.class)
    public boolean setButtonByRoleModule(RoleButtonReq req) {
        List<Long> menuIdList = gihonMenuService.getAllMenu(req.getModuleId(), true).stream().map(m -> m.getId())
            .collect(Collectors.toList());
        List<Long> allbtnList = gihonButtonService.getMenuButton(menuIdList, true).stream().map(m -> m.getId())
            .collect(Collectors.toList());
        List<Long> btnoldList = gihonRoleButtonService.getButtonListByRole(req.getRoleId());
        btnoldList.retainAll(allbtnList);// 求交集,属于当前模块的角色配置
        req.getButtonList().retainAll(allbtnList);
        boolean flag = true;
        if (req.getButtonList().isEmpty()) {
            flag = gihonRoleButtonService.removeButtonByRoleId(req.getRoleId(), btnoldList);
        } else if (btnoldList.isEmpty()) {
            flag = gihonRoleButtonService.insertRoleButton(req.getRoleId(), req.getButtonList());
        } else {
            List<Long> tempList = new ArrayList<>();
            tempList.addAll(req.getButtonList());
            tempList.removeAll(btnoldList);// 新增
            btnoldList.removeAll(req.getButtonList());// 需要删除的
            gihonRoleButtonService.insertRoleButton(req.getRoleId(), tempList);
            gihonRoleButtonService.removeButtonByRoleId(req.getRoleId(), btnoldList);
        }
        CustomerEvent event = new CustomerEvent(req.getModuleId());
        event.setType(3);
        event.setId(req.getModuleId());
        customePushlisher.publish(event);
        return flag;
    }

    @Override
    public List<SelectVal> labelAndValue(Long companyId, String roleName) {
        return this.lambdaQuery().like(StringUtils.hasText(roleName), GihonRole::getRoleName, roleName)
            .eq(companyId != null, GihonRole::getId, companyId).orderByDesc(GihonRole::getRoleName).list().stream()
            .map(c -> SelectVal.builder().value(c.getId().toString()).label(c.getRoleName()).build())
            .collect(Collectors.toList());
    }

}
