package com.gihon.component.rbac.service.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import javax.annotation.PostConstruct;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.stereotype.Service;

import com.fasterxml.jackson.core.type.TypeReference;
import com.gihon.component.entity.BaseEntity;
import com.gihon.component.enums.ModuleType;
import com.gihon.component.properties.RedisConstants;
import com.gihon.component.rbac.entity.GihonButton;
import com.gihon.component.rbac.entity.GihonMenu;
import com.gihon.component.rbac.entity.GihonModule;
import com.gihon.component.rbac.entity.GihonRole;
import com.gihon.component.rbac.entity.GihonUser;
import com.gihon.component.rbac.service.GihonButtonService;
import com.gihon.component.rbac.service.GihonMenuService;
import com.gihon.component.rbac.service.GihonModuleService;
import com.gihon.component.rbac.service.GihonRoleButtonService;
import com.gihon.component.rbac.service.GihonRoleService;
import com.gihon.component.rbac.service.GihonUserService;
import com.gihon.component.rbac.service.PermissionService;
import com.gihon.component.util.JacksonJsonUtils;
import com.gihon.component.web.auth.AuthConstans;
import com.gihon.component.web.auth.AuthUser;
import com.gihon.component.web.auth.AuthUtils;
import com.gihon.component.web.service.TokenService;
import com.gihon.component.web.vo.MenuVO;
import com.gihon.component.web.vo.PermissionVO;

import lombok.extern.slf4j.Slf4j;

@Slf4j
@Service
public class PermissionServiceImpl implements PermissionService {

    @Autowired
    private StringRedisTemplate stringRedisTemplate;

    @Autowired
    private GihonModuleService gihonModuleService;

    @Autowired
    private GihonMenuService gihonMenuService;

    @Autowired
    private GihonButtonService gihonButtonService;

    @Autowired
    private GihonRoleButtonService gihonRoleButtonService;

    @Autowired
    private GihonRoleService gihonRoleService;

    @Autowired
    private GihonUserService gihonUserService;

    @Override
    public List<MenuVO> getPermission(String moduleCode, int moduleType) {

        String menuJson = (String)stringRedisTemplate.opsForHash().get(RedisConstants.MENU_TREE_REDIS,
            moduleCode + RedisConstants.SEP + moduleType);
        TypeReference<List<MenuVO>> valueTypeRef = new TypeReference<List<MenuVO>>() {};
        List<MenuVO> root = JacksonJsonUtils.readObject(menuJson, valueTypeRef);
        // 原始
        List<MenuVO> list = new ArrayList<>(root.size());

        AuthUser user = AuthUtils.getUser();
        if (AuthConstans.SUPER_ADMIN.equals(user.getUsername()) && moduleType == ModuleType.WEB.getCode()) {

            return root;
        }

        // 获取全部的角色
        List<String> role = user.getRoleList();
        // 返回有权限的终极菜单。其他不需要后台控制的菜单有前端自行展示
        cloneCopyMenu(root, list, role);

        return list;
    }

    /**
     * 递归查询菜单Copy
     * 
     * @param vo
     * @param list
     */
    private void cloneCopyMenu(List<MenuVO> root, List<MenuVO> list, List<String> role) {
        if (role == null || role.isEmpty()) {
            // 无角色就无权限
            return;
        }
        for (MenuVO voi : root) {
            List<PermissionVO> permissions = voi.getPermissions();
            if (permissions == null || permissions.isEmpty()) {
                continue;
            }
            List<PermissionVO> pt = permissions.stream().filter(poi -> {
                if (poi.getRoleList() == null || poi.getRoleList().isEmpty()) {
                    // 未配置角色就是只有超级管理员有权限@See AuthServiceImpl#checkPermission
                    return false;
                } else if (poi.getRoleList().stream().anyMatch(role::contains)) {// 角色是任意一个即可
                    return true;
                }
                return false;
            }).collect(Collectors.toList());

            if (!pt.isEmpty()) {// 菜单需要有任意一个操作权限
                MenuVO mv = new MenuVO();
                mv.setMenuCode(voi.getMenuCode());
                mv.setMenuId(voi.getMenuId());
                mv.setPermissions(pt);
                list.add(mv);
            }
        }
    }

    /**
     * 删除用户信息缓存
     */
    @Override
    public void refreshUser(Long id) {
        GihonUser user = gihonUserService.getById(id);
        stringRedisTemplate.opsForHash().delete(TokenService.USER_STORE, user.getUsername());
    }

    @Override
    public void refreshRole(Long id) {
        this.init();
    }
    @Override
    public void refreshModule(Long id) {
        GihonModule module = gihonModuleService.getById(id);
        this.setModule(module);
    }

    private void setModule(GihonModule module) {
        List<GihonMenu> menuList = gihonMenuService.getAllMenu(module.getId(), false);

        List<MenuVO> menus = new ArrayList<>();

        MenuVO root = new MenuVO();
        root.setMenuCode(module.getModuleCode());
        root.setMenuId(-1L);

        menus.add(root);
        MenuVO mp = new MenuVO();

        menuList.forEach(m -> {
            MenuVO mv = new MenuVO();
            mv.setMenuCode(m.getMenuCode());
            mv.setMenuId(m.getId());
            menus.add(mv);
        });
        List<GihonButton> buttonList = gihonButtonService
            .getMenuButton(menuList.stream().map(BaseEntity::getId).collect(Collectors.toList()), false);
        List<PermissionVO> plist = new ArrayList<>();
        for (GihonButton btn : buttonList) {
            List<Long> rbtnList = gihonRoleButtonService.getRoleListByBtn(btn.getId());
            PermissionVO pvo = new PermissionVO();
            pvo.setUrl(btn.getPath());
            pvo.setMethod(btn.getMethod());
            pvo.setBtnCode(btn.getButtonCode());
            pvo.setBtnName(btn.getButtonName());
            if (!rbtnList.isEmpty()) {
                List<GihonRole> rl = gihonRoleService.getRoleListByIds(rbtnList);
                pvo.setRoleList(rl.stream().map(GihonRole::getRoleCode).collect(Collectors.toList()));
            } else {
                pvo.setRoleList(new ArrayList<>(0));
            }
            mp.setMenuId(btn.getMenuId());
            int index = menus.indexOf(mp);
            if (index > 0) {
                menus.get(index).addPermission(pvo);
            }
            if (StringUtils.isBlank(btn.getPath())) {
                continue;
            }
            plist.add(pvo);
        }
        // 用于拦截器的访问权限按钮
        stringRedisTemplate.opsForHash().put(RedisConstants.PERMISSN_REDIS, module.getModuleCode(),
            JacksonJsonUtils.writeObject(plist));
        // 用于拦截器的访问权限菜单
        stringRedisTemplate.opsForHash().put(RedisConstants.MENU_TREE_REDIS, module.getModuleCode(),
            JacksonJsonUtils.writeObject(menus));
    }

    // 定时刷新或者AOP通知
    @PostConstruct
    public void init() {
        try {

            List<GihonModule> moduleList = gihonModuleService.lambdaQuery().list();
            for (GihonModule module : moduleList) {
                List<GihonMenu> menuList = gihonMenuService.getAllMenu(module.getId(), false);

                List<MenuVO> menus = new ArrayList<>();

                MenuVO root = new MenuVO();
                root.setMenuCode(module.getModuleCode());
                root.setMenuId(-1L);

                menus.add(root);
                MenuVO mp = new MenuVO();

                menuList.forEach(m -> {
                    MenuVO mv = new MenuVO();
                    mv.setMenuCode(m.getMenuCode());
                    mv.setMenuId(m.getId());
                    menus.add(mv);
                    // mp.setMenuId(m.getParentId());
                    // int index = menus.indexOf(mp);
                    // if (index > 0) {
                    // menus.get(index).addMenu(mp);
                    // }
                });
                List<GihonButton> buttonList = gihonButtonService
                    .getMenuButton(menuList.stream().map(BaseEntity::getId).collect(Collectors.toList()), false);
                List<PermissionVO> plist = new ArrayList<>();
                for (GihonButton btn : buttonList) {
                    List<Long> rbtnList = gihonRoleButtonService.getRoleListByBtn(btn.getId());
                    PermissionVO pvo = new PermissionVO();
                    pvo.setUrl(btn.getPath());
                    pvo.setMethod(btn.getMethod());
                    pvo.setBtnCode(btn.getButtonCode());
                    pvo.setBtnName(btn.getButtonName());
                    if (!rbtnList.isEmpty()) {
                        List<GihonRole> rl = gihonRoleService.getRoleListByIds(rbtnList);
                        pvo.setRoleList(rl.stream().map(r -> r.getCompanyId() + RedisConstants.SEP + r.getRoleCode())
                            .collect(Collectors.toList()));
                    } else {
                        pvo.setRoleList(new ArrayList<>(0));
                    }
                    mp.setMenuId(btn.getMenuId());
                    int index = menus.indexOf(mp);
                    if (index > 0) {
                        menus.get(index).addPermission(pvo);
                    }
                    if (StringUtils.isBlank(btn.getPath())) {
                        continue;
                    }
                    plist.add(pvo);
                }
                // 用于拦截器的访问权限按钮
                stringRedisTemplate.opsForHash().put(RedisConstants.PERMISSN_REDIS,
                    module.getModuleCode() + RedisConstants.SEP + module.getModuleType(),
                    JacksonJsonUtils.writeObject(plist));
                // 用于拦截器的访问权限按钮
                stringRedisTemplate.opsForHash().put(RedisConstants.MENU_TREE_REDIS,
                    module.getModuleCode() + RedisConstants.SEP + module.getModuleType(),
                    JacksonJsonUtils.writeObject(menus));
            }
        } catch (Exception e) {
            log.error("初始化模块菜单失败:", e);
        }
    }

}
